# Software for collecting and servicing data for smart houses
This software was created to run on a Raspberry Pi within a house that uses IoT devices to control heat pumps through a smart algorithm.

### Latest updates:
- MET have replaced Open Weather Map for supplying outside temperature to the heating algorithm (MHE/MPC script). Open Weather Map can still collect data, but is now optional.
- **/api/get_data_since** API endpoint is added. This endpoint makes it easier for the MHE/MPC script to only request the data since the last iteration.
- The **< date_format >** parameter for the endpoints /api/get_data and /api/get_file_list now only supports 'daily'. Weekly, monthly and yearly data files was removed due to performance issues

## Features
- Collects data from tibber, sensibo, MET.no and Open Weather Map
- Provides a web API to request data and managing the data collection
- Backups all the data every day at midnight
- With the help of Docker, the data collector restarts automatically if there are any crashes

## Installing and Running the software
See [installation instructions](docs/installation.md).

## API Endpoints
Once the software started on the Raspberry Pi, the following API endpoints can be used:

### Base IP address
The pi can be accessed locally by the address (try in the web browser):
- raspberrypi.local

The actual IP address can be found by:
- $ ping raspberrypi.local (from your other computer)

### Static files
- {ip_address}/data/
  - .pkl data files
- {ip_address}/backups/
  - .zip Backups of /data:
- {ip_address}/readme
  - Readme.md (this file)

### Data collector API
**Start the data collector** by sending a POST request (responds with json status, message, and params)
- POST http://{ip_address}/api/start_data_collector
  - Note: The params are _optional_. If no params are supplied in the POST request, the data collector will start with the last-used settings. If there are no last-used settings, and no params supplied, the data_collector will not start.
  - example:
    ```python
    import requests
    
    response = requests.post('http://{ip_address}/api/start_data_collector', json=params)

    print(response.text)
    ```
  - with supplied parameters:
    ```python
    # Params are optional. Will use last-used params if none are supplied in the post request.
    # However, if you do supply params, do it in the following structure:
    params = { 
        'tibber': {
            'api_key': 'api-key',
            'sampling_time': 60, # [minutes] - Optional
            'rt_sampling_time': 20 # [seconds] - Optional
        },
        'sensibo': {
            'api_key': 'api-key',
            'sampling_time': 5, # [minutes] - Optional
        },
        'MET': { # Optional (will use default values from Config, if no params are supplied)
            'lat': "63.4",
            'lon': "10.335",
            'altitude': 200,
            'location_name': 'ugla',
            'personal_id': 'sebastien.gros@ntnu.no ericth@stud.ntnu.no POWIOT project'
        },
        'open_weather_map': { # Optional
            'api_key': 'api-key',
            'lat': "63.4",
            'lon': "10.335",
            'sampling_time': 60, # minutes
        },
        'backup': { # Optional
            'run_backups': True, # Optional (Defaults to True)
        },
    }
    ```

**Stop the data collector** (returns JSON message informing you on success/fail)
- POST http://{ip_address}/api/stop_data_collector
  - example:
  ```python
    import requests
    
    response = requests.post('http://{ip_address}/api/stop_data_collector')

    print(response.text)
    ```

**Get file list** (returns JSON)
- GET http://{ip_address}/api/get_file_list/< sensor >/< date_format >/< date >
    - < sensor > 
        - 'tibber'
        - 'tibber-realtime-home-pumps'
        - 'tibber-realtime-home-up'
        - 'sensibo'
        - 'open_weather_map'
        - 'MET'
    - < date_format >
        - 'daily'
    - < date > 
        - 'first' (will get the name of the first file that was created)
        - 'latest' (will get the name of the latest file that was created)
        - 'all' (will get a list of the names of all the data files)
        - '2020-09-02' (will get a list of the names of the data files from a single day)
    ```python
    import requests
    import json

    response = requests.get('http://{ip_address}/api/get_file_list/sensibo/daily/all')

    files_list = json.loads(response.text)
    ```

**Get data** (returns JSON)
- GET http://{ip_address}/api/get_data/< sensor >/< date_format >/< date >
    - Parameters are the same as for _Get file list_, but:
    - < date >:
        - 'first' (will get the data from the first data file that was created.)
        - 'latest' (will get the data from the latest data file that was created.)
        - 'all' (will get all the data from all data files.)
        - '2020-09-02' (will get all data from a specific day.)
    ```python
    import requests
    import json

    response = requests.get('http://{ip_address}/api/get_data/MET/daily/2020-09-02')

    response = requests.get('http://{ip_address}/api/get_data/MET/daily/all')

    response = requests.get('http://{ip_address}/api/get_data/MET/daily/latest')

    response = requests.get('http://{ip_address}/api/get_data/MET/daily/first')

    data = json.loads(response.text)
    ```
    
**Get data since a specified time** (returns JSON)
- GET http://{ip_address}/api/get_data_since/< sensor >/< time_type >/< time >
    - < sensor >
    - < time_type >
        - 'hours'
        - 'minutes'
        - 'timestamp'
    - < time >:
        - e.g. 7 (any positive integer, if < time_type > is set to 'hours' or 'minutes')
        - e.g. 'Wed, 09 Jun 2021 11:49:09 GMT' (a string timestamp in this format, if <time_type> is set to 'timestamp')
    ```python
    import requests
    import json

    response = requests.get('http://{ip_address}/api/get_data_since/sensibo/hours/5')
    
    response = requests.get('http://{ip_address}/api/get_data_since/sensibo/minutes/5')

    response = requests.get('http://{ip_address}/api/get_data_since/sensibo/timestamp/Wed, 09 Jun 2021 11:49:09 GMT')

    data = json.loads(response.text)
    ```

## Data collection not working?
1) Check the API keys!
2) Check **log__data_collector.log** for error messages.