import logging
from uuid import uuid1
from pathlib import Path
import pickle
import numpy as np
import threading
from queue import Queue
from datetime import datetime, timedelta, timezone
import time
import pytz
import tibber
import asyncio
import requests
import json
import os
from metno_locationforecast import Place, Forecast
from dotenv import load_dotenv
load_dotenv()

import data_collector.sensibo_client as SC
import data_collector.helpers as helpers
import data_collector.backup as backup
from data_collector.config import Config

logging.basicConfig(filename='log__data_collector.log',
                    filemode='a',
                    format='%(asctime)s,%(msecs)d %(name)s %(levelname)s %(message)s',
                    datefmt='%Y-%m-%d--%H:%M:%S',
                    level=logging.WARNING)

#########################################################################################
#
# This data collector package is based on a producer and consumer system, with loops in different threads: 
# - one thread for each "sensor" (API),
# - one thread for the backup function, 
# - one main handler thread.
#
# The main handler thread receives data and requests from all other threads through a FIFO queue.
# Data requests from the web interface also sends requests to this queue. How the main handler thread
# handles the queue item depends on the message in the queue item. The purpose of having a "main handler"
# is to remove the risk of data races when trying to write to, and access, the data files.
#
# The sensor threads have different sampling rates, which are defined by the parameters that are supplied
# by the used when starting the data collector. At each interval they request data from their respective API,
# then sends this data to the main handler, which stores the data to files.
#
# The data is stored in .pkl files in this folder structure:
# - data/<sensor>
#               /daily
#               /weekly (Deprecated, this folder is not used anymore and should be removed from the software)
#               /monthly (Deprecated, this folder is not used anymore and should be removed from the software)
#               /yearly (Deprecated, this folder is not used anymore and should be removed from the software)
#
# Backups are stored in /backups
#
# The code is structured in the following sections:
# - Sensors (code related to each sensor. Each has a loop which is run in its own thread.)
#       - MET
#       - Tibber realtime
#       - Tibber hourly consumption and cost data
#       - Sensibo
#       - Open Weather Map (This has been "replaced" by MET, but can still be used to collect data.)
# - Request handler functions (requests that comes from a user while the data_collector is running)
# - Backup loop (run in its own thread)
# - Main handler loop (run in its own thread)
# - Start function (Calling this will start the data_collector)
# - Stop function (Calling this will stop the data_collector)
#
#
# TO ADD A SENSOR, FOLLOW THESE STEPS:
# 0) First understand how the code works for the other sensors
# 1) Add sensor class and file saving functions (see other sensors)
# 2) Add the sensor to the main-handler loop (line 896)
#   - Make sure to specify a sensor_name in the sensor class, and use this in the main handler thread
# 3) Add the sensor to the start function (line 956)
#   - Get parameters for this sensor from the "params" argument (line 968 and onwards)
#   - Make sure a sensor object is initialized  (line 1081 and onwards) 
#   - Initialize and start a thread with the sensors mining loop (see others for example) (line 1135)
#   - Add the sensor_name (must be the same) to the list in generate_all_data_folders call (line 1079)
# 4) Add the sensor to helpers.write_to_env_file() (line 16 in helpers.py)
# 5) Add default parameters to config.py
#
########################################################################################


########################################################################################
################################ Solar Radiation forecast ##############################
# TODO


########################################################################################
################################ MET weather forecast ##################################
# https://api.met.no/doc/TermsOfService
# * Anything over 20 requests/second per application (total, not per client) requires special agreement.
# * Do not ask too often, and don't repeat requests until the time indicated in the Expires response header.
# * Cache data locally and and use the If-Modified-Since request header to avoid repeatedly downloading the same data.
# * Don't schedule many requests at the same time, e.g. every hour on the dot or when the forecast model runs are finished. 
# Add a random number of minutes to the time of the requests as our data are continously updated. 
# Spread your traffic evenly out over time so it makes a flat curve, not a sawtooth.
# * When using requests with latitude/longitude, truncate all coordinates to max 4 decimals. 
# There is no need to ask for weather forecasts with nanometer precision! For new products, requests with 5+ decimals will return a 403 Forbidden.
#
# Resources: https://pypi.org/project/metno-locationforecast/


class METForecast():
    def __init__(self, 
                 lat            = Config.DEFAULT_PARAMS["MET"]["lat"], 
                 lon            = Config.DEFAULT_PARAMS["MET"]["lon"], 
                 alti           = Config.DEFAULT_PARAMS["MET"]["altitude"], 
                 location_name  = Config.DEFAULT_PARAMS["MET"]["location_name"], 
                 personal_id    = Config.DEFAULT_PARAMS["MET"]["personal_id"]
                 ):
        self._lat = lat
        self._lon = lon
        self._alti = alti
        self._personal_id = personal_id
        self._place = Place(location_name, lat, lon, alti) # etter tråden startets?
        self._forecast = Forecast(self._place, self._personal_id, forecast_type='compact', save_location='./data/MET/temp')
        self._timezone = Config.LOCAL_TIMEZONE
        self._sensor_name = 'MET'
        self._session_id = str(int(time.time())) + '-' + str(uuid1())[:8]
        self._last_data_point = {
            'time': datetime.now(tz=Config.LOCAL_TIMEZONE),
            'temperature': 999999999
        }

    # This method is called in the loop within start_mining_loop
    def _get_latest_data(self, forecast, local_tz):
        update = forecast.update() # update becomes one of these: 'Data-Not-Expired', 'Data-Not-Modified' or 'Data-Modified'
        if update != 'Data-Modified' and self._last_data_point['temperature'] != 999999999: 
            time_now_string = datetime.now(tz=Config.LOCAL_TIMEZONE).strftime('%Y-%m-%d--%H:%M:%S')
            print(f"{time_now_string} - MET      --->      No new data since last update")
            print(f"{time_now_string} - MET      ---->     The current data header expires at: {vars(self._forecast.data)['expires'].replace(tzinfo=timezone.utc).astimezone(self._timezone)}")
            return None
        
        data = vars(forecast.data)['intervals']
        met_weather_data = {
            'time'        : [],
            'temperature' : [],
            'sampling_time' : []
        }

        # Get timestamp and temperature of latest data_point
        timestamp = data[0].start_time.replace(tzinfo=timezone.utc).astimezone(local_tz)
        temperature = data[0].variables['air_temperature'].value

        met_weather_data['time'].append(timestamp)
        met_weather_data['temperature'].append(temperature)
        sampling_time = (timestamp - self._last_data_point['time']).total_seconds()/60
        if sampling_time <= 0:
            sampling_time = None
        met_weather_data['sampling_time'].append(sampling_time)
        
        if not self._last_data_point \
        or timestamp > self._last_data_point['time'] \
        or temperature != self._last_data_point['temperature']:
            self._last_data_point['time'] = timestamp
            self._last_data_point['temperature'] = temperature
            return met_weather_data
        else:
            return None

    # This method is called when the thread is started, and starts a loop
    def start_mining_loop(self, main_q, stop_q, scheduled_time=None):
        time_now_string = datetime.now(tz=Config.LOCAL_TIMEZONE).strftime('%Y-%m-%d--%H:%M:%S')
        print(f"{time_now_string} - {self._sensor_name} started mining")

        if not scheduled_time:
            scheduled_time = datetime.now(tz=Config.LOCAL_TIMEZONE) + timedelta(minutes=5)

        while True:
            try:
                message = stop_q.get(block=False)
                if message is _sentinel:
                    stop_q.put(_sentinel)
                    time_now_string = datetime.now(tz=Config.LOCAL_TIMEZONE).strftime('%Y-%m-%d--%H:%M:%S')
                    print(f"{time_now_string} - Stopped: " + self._sensor_name)
                    break
            except:
                pass
            helpers.sleep_until(scheduled_time)

            data = self._get_latest_data(self._forecast, self._timezone)
            if data:
                message = (self._sensor_name, data, self._session_id)
                main_q.put(message) # send data to handler_thread

            scheduled_time = vars(self._forecast.data)['expires'].replace(tzinfo=timezone.utc).astimezone(self._timezone)
            while scheduled_time < datetime.now(tz=Config.LOCAL_TIMEZONE):
                scheduled_time = scheduled_time + timedelta(minutes=5)

# This function is used by save_met_data_to_all_periodic_files()
def save_met_data_to_file(filename, data_point):
    try:
        file = Path(filename)
        # If file exists, append to it
        if file.is_file():
            with open(filename, 'rb') as f:
                file_data = pickle.load(f)
                file_data['time'].append(data_point['time'][0])
                file_data['temperature'].append(data_point['temperature'][0])
                file_data['sampling_time'].append(data_point['sampling_time'][0])
            with open(filename, 'wb') as f:
                pickle.dump(file_data, f)
        # Else, write to new file
        else:
            if data_point:
                with open(filename, 'wb') as f:
                    pickle.dump(data_point, f)
    except Exception as e:
        time_now_string = datetime.now(tz=Config.LOCAL_TIMEZONE).strftime('%Y-%m-%d--%H:%M:%S')
        print(f"{time_now_string} - something went wrong writing MET Forecast data to file: ", filename)
        print(str(e))
        return False
    return True

# This function is used by the main handler thread to store the data to pickle files
def save_met_data_to_all_periodic_files(data_point, session_id, time_now):
    # Try to get the actual time for the datapoint, else use the current time from the argument
    try:
        time_now = data_point['time'][0]
    except:
        pass
    # Create a list of file names to use when saving the data
    fname_list = helpers.get_sensor_data_file_names('MET', session_id, time_now)
    success_flags = []

    for fname in fname_list:
        if not save_met_data_to_file(fname, data_point):
            success_flags.append(False)
    
    if False in success_flags:
        time_now_string = datetime.now(tz=Config.LOCAL_TIMEZONE).strftime('%Y-%m-%d--%H:%M:%S')
        print(f"{time_now_string} - One or more of the MET Forecast file saves failed.")
        return False
    return True



########################################################################################

########################################################################################
################################ Tibber Realtime sensor ################################
# Lowest sampling rate is 2 seconds
#
# Resources:
# - https://pypi.org/project/pyTibber/
# - https://github.com/Danielhiversen/pyTibber
# - https://developer.tibber.com/

class TibberRT():
    """
    Tibber realtime sensor
    * Creates a continous subscription which sends live measured data every 2 seconds
    """
    def __init__(self, api_key, sensor_name, sensor_number, sampling_time=20):
        self._api_key = api_key
        self._session_id = str(int(time.time())) + '-' + str(uuid1())[:8]
        self._sensor_name = sensor_name
        self._sensor_number = sensor_number
        self._sampling_time = sampling_time

        # Siden det er per Home, så trenger det kanskje være på separate tråder
    
    # To get data from tibber you must first make connection
    def connect_and_get_home(self, sensor_number):
        try:
            self._tibber_conn = tibber.Tibber(
                access_token=self._api_key
            )
            self._tibber_conn.sync_update_info()
            
            self._homes = self._tibber_conn.get_homes()
            return self._homes[sensor_number]
        except Exception as e:
            time_now_string = datetime.now(tz=Config.LOCAL_TIMEZONE).strftime('%Y-%m-%d--%H:%M:%S')
            logging.error("Realtime Tibber: could not connect: " + str(e))
            print(f"{time_now_string} - Realtime Tibber: could not connect.\n")
            return None

    def disconnect(self):
        self._tibber_conn.sync_close_connection()

    # An async callback function must be passed to the real-time subscription
    async def _async_callback(self, payload=None):
        """Handle received data, then pass to passed callback function"""
        time_now = datetime.now(tz=Config.LOCAL_TIMEZONE)

        if time_now < self._next_rt_measurement_time:
            return

        errors = payload.get("errors")
        if errors:
            logging.error(errors[0])
            return
        data = payload.get("data")
        if data is None:
            return
        live_measurement = data.get("liveMeasurement")
        if live_measurement is None:
            return

        self._realtime_callback(live_measurement)

    # This function is called by the async callback function and gets the real-time data
    def _realtime_callback(self, live_measurement):
        try:
            message = stop_q.get(block=False)
            if message is _sentinel:
                stop_q.put(_sentinel)
                time_now_string = datetime.now(tz=Config.LOCAL_TIMEZONE).strftime('%Y-%m-%d--%H:%M:%S')
                print(f"{time_now_string} - Stopped: " + self._sensor_name)
                try:
                    self._tibber_conn.sync_close_connection()
                    self._loop.stop()
                    return
                except:
                    self._loop.stop()
                    return
        except:
            pass
        t_stamp_split = live_measurement['timestamp'].split('+')
        GMT_split = t_stamp_split[1].split(':')
        time_str = t_stamp_split[0] + '+' + GMT_split[0] + GMT_split[1]
        timestamp = datetime.strptime(time_str,'%Y-%m-%dT%H:%M:%S%z')
        timestamp_local = timestamp.astimezone(tz=Config.LOCAL_TIMEZONE)

        data = {}

        # take data and send it to main handler
        data['time'] = [timestamp_local]
        data['sampling_time'] = [self._sampling_time]
        data['power'] = [live_measurement['power']]
        data['accumulatedConsumption'] = [live_measurement['accumulatedConsumption']] # Since midnight kWh
        data['accumulatedCost'] = [live_measurement['accumulatedCost']] # since midnight

        if type(data['sampling_time']) == tuple:
            data['sampling_time'] = data['sampling_time'][0]

        if data:
            message = (self._sensor_name, data, self._session_id)
            self._main_q.put(message) # send data to handler_thread
        
        self._next_rt_measurement_time += timedelta(seconds=self._sampling_time)

    # This function is called when the real-time tibber thread starts
    def start_rt(self, main_q, stop_q):
        self._main_q = main_q
        self._stop_q = stop_q

        self._loop = asyncio.new_event_loop()
        asyncio.set_event_loop(self._loop)
        
        self._home = self.connect_and_get_home(self._sensor_number)
        while self._home == None:
            time_now_string = datetime.now(tz=Config.LOCAL_TIMEZONE).strftime('%Y-%m-%d--%H:%M:%S')
            print(f"{time_now_string} - {self._sensor_name}: Retrying to connect to tibber in five minutes")
            time.sleep(300)
            self._home = self.connect_and_get_home(self._sensor_number)
        
        self._home.sync_update_info()
        if self._home.has_real_time_consumption:
            time_now_string = datetime.now(tz=Config.LOCAL_TIMEZONE).strftime('%Y-%m-%d--%H:%M:%S')
            print(f"{time_now_string} - {self._sensor_name} started mining | sampling time: {self._sampling_time} sec")
            self._next_rt_measurement_time = datetime.now(tz=Config.LOCAL_TIMEZONE) + timedelta(seconds=self._sampling_time)
            self._loop.create_task(self._home.rt_subscribe(self._loop, self._async_callback))
            self._loop.run_forever()
        else:
            print(f"{time_now_string} - {self._sensor_name} had no real_time_subscription, stopping thread")
            return

# Used in save_tibber_realtime_data_to_all_periodic_files()
def save_tibber_realtime_data_to_file(filename, data_point):
    try:
        file = Path(filename)
        # If file exists, append to it
        if file.is_file():
            with open(filename, 'rb') as f:
                file_data = pickle.load(f)
                file_data['time'].append(data_point['time'][0])
                file_data['sampling_time'].append(data_point['sampling_time'][0])
                file_data['power'].append(data_point['power'][0])
                file_data['accumulatedConsumption'].append(data_point['accumulatedConsumption'][0])
                file_data['accumulatedCost'].append(data_point['accumulatedCost'][0])
            with open(filename, 'wb') as f:
                pickle.dump(file_data, f)
        # Else, write to new file
        else:
            if data_point:
                with open(filename, 'wb') as f:
                    pickle.dump(data_point, f)
    except Exception as e:
        time_now_string = datetime.now(tz=Config.LOCAL_TIMEZONE).strftime('%Y-%m-%d--%H:%M:%S')
        print(f"{time_now_string} - something went wrong writing tibber data to file: ", filename)
        print(str(e))
        return False
    return True

# Used by the main handler to store data to files for tibber
def save_tibber_realtime_data_to_all_periodic_files(data_point, session_id, sensor, time_now):
    # Try to get the actual time for the datapoint, else use the current time from the argument
    try:
        time_now = data_point['time'][0]
    except:
        pass

    # Create a list of file names to use when saving the data
    fname_list = helpers.get_sensor_data_file_names(sensor, session_id, time_now)
    success_flags = []

    for fname in fname_list:
        if not save_tibber_realtime_data_to_file(fname, data_point):
            success_flags.append(False)
    
    if False in success_flags:
        time_now_string = datetime.now(tz=Config.LOCAL_TIMEZONE).strftime('%Y-%m-%d--%H:%M:%S')
        print(f"{time_now_string} - One or more of the realtime tibber file saves failed.")
        return False
    return True



########################################################################################
################################ Tibber hourly data ####################################
# Resources:
# - https://pypi.org/project/pyTibber/
# - https://github.com/Danielhiversen/pyTibber
# - https://developer.tibber.com/

class TibberAPI():
    def __init__(self, api_key, sampling_time=None):
        self._api_key = api_key
        self._session_id = str(int(time.time())) + '-' + str(uuid1())[:8]
        self._sampling_time = sampling_time if int(sampling_time) else Config.DEFAULT_PARAMS["tibber"]["sampling_time"]
        self._sensor_name = "tibber"

    # Used to make a connection to the Tibber API
    def connect_and_get_homes(self):
        try:
            self._tibber_conn = tibber.Tibber(
                access_token=self._api_key,
                timeout=600
            )
            self._tibber_conn.sync_update_info()
            return self._tibber_conn.get_homes()
        except Exception as e:
            time_now_string = datetime.now(tz=Config.LOCAL_TIMEZONE).strftime('%Y-%m-%d--%H:%M:%S')
            logging.error("Tibber Hourly: could not connect: " + str(e))
            print(f"{time_now_string} - Tibber Hourly: could not connect.\n")
            print(e)
            return None

    # Called when starting the Tibber hourly thread
    def start_mining_loop(self, main_q, stop_q, scheduled_time, sampling_time):
        time_now_string = datetime.now(tz=Config.LOCAL_TIMEZONE).strftime('%Y-%m-%d--%H:%M:%S')
        print(f"{time_now_string} - {self._sensor_name} started mining | sampling time: {sampling_time} min | next scheduled measurement: {scheduled_time}")

        self._loop = asyncio.new_event_loop()
        asyncio.set_event_loop(self._loop)

        self._homes = self.connect_and_get_homes()

        while True:
            try:
                message = stop_q.get(block=False)
                if message is _sentinel:
                    stop_q.put(_sentinel)
                    time_now_string = datetime.now(tz=Config.LOCAL_TIMEZONE).strftime('%Y-%m-%d--%H:%M:%S')
                    print(f"{time_now_string} - Stopped: " + self._sensor_name)
                    try:
                        self._tibber_conn.sync_close_connection()
                        break
                    except:
                        break
            except:
                pass

            helpers.sleep_until(scheduled_time)
            if not self._homes:
                self._tibber_conn.sync_close_connection()
                self._homes = self.connect_and_get_homes()
            data = self._get_latest_measurement(homes=self._homes)
            if data:
                message = ("tibber", data, self._session_id)
                main_q.put(message) # send data to handler_thread

            scheduled_time += timedelta(minutes=sampling_time)

    def _get_latest_measurement(self, homes=[None]):
        data = {}
        resolution = "HOURLY"
        n = 10
        
        for home in homes:
            try:
                home.sync_update_info()
                home_name = home.info['viewer']['home']['appNickname']
                data[home_name] = {}
                
                historic_data = home.sync_get_historic_data(n, resolution)
                for data_point_ in reversed(historic_data):
                    if data_point_['consumption']:
                        data_point = data_point_
                        break

                t_stamp_split = data_point['from'].split('+')
                GMT_split = t_stamp_split[1].split(':')
                time_str = t_stamp_split[0] + '+' + GMT_split[0] + GMT_split[1]
                timestamp = datetime.strptime(time_str,'%Y-%m-%dT%H:%M:%S%z')
                timestamp_local = timestamp.astimezone(tz=Config.LOCAL_TIMEZONE)

                data[home_name]['time'] = [timestamp_local]
                data[home_name]['sampling_time'] = [self._sampling_time],
                data[home_name]['consumption'] = [data_point['consumption']]
                data[home_name]['cost'] = [data_point['cost']]
                data[home_name]['total_cost'] = [data_point['totalCost']]
                
                if type(data[home_name]['sampling_time']) == tuple:
                    data[home_name]['sampling_time'] = data[home_name]['sampling_time'][0]
            except Exception as e:
                time_now_string = datetime.now(tz=Config.LOCAL_TIMEZONE).strftime('%Y-%m-%d--%H:%M:%S')
                logging.error("Tibber: could not get historic data. " + str(e))
                print(f"{time_now_string} - Tibber: could not get historic data.")
                print(e)
                return False
        return data

# Used by save_tibber_data_to_all_periodic_files()
def save_tibber_data_to_file(filename, data_point):
    try:
        file = Path(filename)
        # If file exists, append to it
        if file.is_file():
            with open(filename, 'rb') as f:
                file_data = pickle.load(f)
                for home_name in data_point:
                    file_data[home_name]['time'].append(data_point[home_name]['time'][0])
                    file_data[home_name]['sampling_time'].append(data_point[home_name]['sampling_time'][0])
                    file_data[home_name]['consumption'].append(data_point[home_name]['consumption'][0])
                    file_data[home_name]['cost'].append(data_point[home_name]['cost'][0])
                    file_data[home_name]['total_cost'].append(data_point[home_name]['total_cost'][0])
            with open(filename, 'wb') as f:
                pickle.dump(file_data, f)
        # Else, write to new file
        else:
            if data_point:
                with open(filename, 'wb') as f:
                    pickle.dump(data_point, f)
    except Exception as e:
        time_now_string = datetime.now(tz=Config.LOCAL_TIMEZONE).strftime('%Y-%m-%d--%H:%M:%S')
        print(f"{time_now_string} - something went wrong writing tibber data to file: ", filename)
        print(str(e))
        return False
    return True

# Used by the main handler to store data to files for tibber
def save_tibber_data_to_all_periodic_files(data_point, session_id, time_now):
    # Try to get the actual time for the datapoint, else use the current time from the argument
    for home_name in data_point:
        try:
            time_now = data_point[home_name]['time'][0]
            break
        except:
            continue
    # Create a list of file names to use when saving the data
    fname_list = helpers.get_sensor_data_file_names('tibber', session_id, time_now)
    success_flags = []

    for fname in fname_list:
        if not save_tibber_data_to_file(fname, data_point):
            success_flags.append(False)
    
    if False in success_flags:
        time_now_string = datetime.now(tz=Config.LOCAL_TIMEZONE).strftime('%Y-%m-%d--%H:%M:%S')
        print(f"{time_now_string} - One or more of the tibber file saves failed.")
        return False
    return True


########################################################################################
################################ Sensibo sensor ########################################
# The measurements updates every 90-91seconds, so no use in having a sampling rate shorter than this
#
# Resources: https://sensibo.github.io/

class SensiboSensor():
    def __init__(self, api_key, sampling_time=None):
        self._api_key = api_key
        self._session_id = str(int(time.time())) + '-' + str(uuid1())[:8]
        self._sampling_time = sampling_time if int(sampling_time) else Config.DEFAULT_PARAMS["sensibo"]["sampling_time"]
        self._sensor_name = "sensibo"

        self._what_to_measure = ['temperature', 'humidity']
        self._states_to_record = ['on', 'targetTemperature', 'fanLevel', 'mode']

    def connect_and_get_devices(self):
        try:
            self._home = SC.SensiboClientAPI(self._api_key)
            return self._home.devices()
        except Exception as e:
            time_now_string = datetime.now(tz=Config.LOCAL_TIMEZONE).strftime('%Y-%m-%d--%H:%M:%S')
            print(f"{time_now_string} - Home Sensibo could not be accessed.")
            print(str(e))
            return None

    # Called by the sensibo thread when the thread starts
    def start_mining_loop(self, main_q, stop_q, scheduled_time, sampling_time):
        time_now_string = datetime.now(tz=Config.LOCAL_TIMEZONE).strftime('%Y-%m-%d--%H:%M:%S')
        print(f"{time_now_string} - {self._sensor_name} started mining | sampling time: {sampling_time} min | next scheduled measurement: {scheduled_time}")

        self._devices = self.connect_and_get_devices()

        while True:
            try:
                message = stop_q.get(block=False)
                if message is _sentinel:
                    stop_q.put(_sentinel)
                    time_now_string = datetime.now(tz=Config.LOCAL_TIMEZONE).strftime('%Y-%m-%d--%H:%M:%S')
                    print(f"{time_now_string} - Stopped: " + self._sensor_name)
                    break
            except:
                pass

            helpers.sleep_until(scheduled_time)
            if not self._devices:
                self._devices = self.connect_and_get_devices()
            data = self._get_latest_measurement(pumps=self._devices.keys())
            if data:
                message = ("sensibo", data, self._session_id)
                main_q.put(message) # send data to handler_thread
            else:
                time_now_string = datetime.now(tz=Config.LOCAL_TIMEZONE).strftime('%Y-%m-%d--%H:%M:%S')
                print(f"{time_now_string} - Sensibo: Could not get any devices. Continuing to next interval without getting measurement.")

            scheduled_time += timedelta(minutes=sampling_time)

    def _get_latest_measurement(self, pumps=[None]):
        data = {}

        # Avoid keyerrors
        for pump in pumps:
            data[pump] = {
                'measurements': {},
                'states': {}
            }

        for pump in pumps:
            try:
                latest_measurement = self._home.pod_measurement(self._devices[pump])[0]
                current_state = self._home.pod_ac_state(self._devices[pump])
            except Exception as e:
                time_now_string = datetime.now(tz=Config.LOCAL_TIMEZONE).strftime('%Y-%m-%d--%H:%M:%S')
                logging.error("Sensibo get_latest_measurement: Could not get measurements or state")
                print(f"{time_now_string} - Sensibo get_latest_measurement: Could not get measurements or state")
                print(str(e))
                return False
            
            for measurement_type in self._what_to_measure:
                if measurement_type in latest_measurement:
                    data[pump]['measurements'][measurement_type] = [latest_measurement[measurement_type]]
                else:
                    data[pump]['measurements'][measurement_type] = [None]
                    print(latest_measurement)
                    print('Sensibo measurement '+str(measurement_type) +' for '+str(pump)+' missing, put NaN ')

            for state in self._states_to_record:
                    if state in current_state:
                        data[pump]['states'][state] = [current_state[state]]
                    else:
                        data[pump]['states'][state] = [None]
                        print(current_state)
                        print('Sensibo state '+str(state)+' for ' + str(pump)+' missing, put NaN ')
            timestamp = datetime.strptime(latest_measurement['time']['time'], '%Y-%m-%dT%H:%M:%S.%fZ')
            utc_timestamp = timestamp.replace(tzinfo=pytz.utc)
            local_timestamp = utc_timestamp.astimezone(tz=Config.LOCAL_TIMEZONE)
            data[pump]['time'] = [local_timestamp]
            data[pump]['sampling_time'] = [self._sampling_time]
        return data

# Used by save_sensibo_data_to_all_periodic_files()
def save_sensibo_data_to_file(filename, data_point):
    try:
        file = Path(filename)
        if file.is_file():
            with open(filename, 'rb') as f:
                file_data = pickle.load(f)
                for pump in data_point:
                    file_data[pump]['time'].append(data_point[pump]['time'][0])
                    file_data[pump]['sampling_time'].append(data_point[pump]['sampling_time'][0])
                    for measurement in data_point[pump]['measurements'].keys():
                        file_data[pump]['measurements'][measurement].append(data_point[pump]['measurements'][measurement][0])
                    for state in data_point[pump]['states'].keys():
                        file_data[pump]['states'][state].append(data_point[pump]['states'][state][0])
            with open(filename, 'wb') as f:
                pickle.dump(file_data, f)
        else:
            if data_point:
                with open(filename, 'wb') as f:
                    pickle.dump(data_point, f)
    except Exception as e:
        print("something went wrong writing sensibo data to file: ", filename)
        print(str(e))
        return False
    return True

# Used by the main handler thread to store the sensibo data to files
def save_sensibo_data_to_all_periodic_files(data_point, session_id, time_now):
    # Try to get the actual time for the datapoint, else use the current time from the argument
    for pump in data_point:
        try:
            time_now = data_point[pump]['time'][0]
            break
        except:
            continue
    # Create a list of file names to use when saving the data
    fname_list = helpers.get_sensor_data_file_names('sensibo', session_id, time_now)
    success_flags = []

    for fname in fname_list:
        if not save_sensibo_data_to_file(fname, data_point):
            success_flags.append(False)

    if False in success_flags:
        print("One or more of the sensibo file saves failed.")
        return False
    return True


########################################################################################
################################ Open Weather Map API ##################################
# This data is no longer used and has been replaced by the weather data from MET. 
# But it still works.
#
# Resources: https://openweathermap.org/api

class WeatherAPI():
    def __init__(self, api_key, lat, lon, sampling_time=None):
        self._session_id = str(int(time.time())) + '-' + str(uuid1())[:8]
        self._sampling_time = sampling_time if int(sampling_time) else Config.DEFAULT_PARAMS["sensibo"]["sampling_time"]
        self._sensor_name = 'open_weather_map'
        self._url = f"http://api.openweathermap.org/data/2.5/weather?lat={lat}&lon={lon}&appid={api_key}&units=metric"

    def start_mining_loop(self, main_q, stop_q, scheduled_time, sampling_time):
        time_now_string = datetime.now(tz=Config.LOCAL_TIMEZONE).strftime('%Y-%m-%d--%H:%M:%S')
        print(f"{time_now_string} - {self._sensor_name} started mining | sampling time: {sampling_time} min | next scheduled measurement: {scheduled_time}")

        while True:
            try:
                message = stop_q.get(block=False)
                if message is _sentinel:
                    stop_q.put(_sentinel)
                    time_now_string = datetime.now(tz=Config.LOCAL_TIMEZONE).strftime('%Y-%m-%d--%H:%M:%S')
                    print(f"{time_now_string} - Stopped: " + self._sensor_name)
                    break
            except:
                pass

            helpers.sleep_until(scheduled_time)
            data = self._get_latest_measurement(url=self._url)
            if data:
                message = ('open_weather_map', data, self._session_id)
                main_q.put(message) # send data to handler_thread

            scheduled_time += timedelta(minutes=sampling_time)

    def _get_latest_measurement(self, url=None):
        data = {}
        try:
            response = requests.get(url, timeout=300) # timeout after 5 minutes
            if response.status_code == 200:
                current_data = json.loads(response.text)
                timestamp = datetime.utcfromtimestamp(current_data['dt']).replace(tzinfo=timezone.utc)
                local_timestamp = timestamp.astimezone(Config.LOCAL_TIMEZONE)
                
                data['time'] = [local_timestamp]
                data['sampling_time'] = [self._sampling_time]
                data['temperature'] = [current_data['main']['temp']]
            else: 
                time_now_string = datetime.now(tz=Config.LOCAL_TIMEZONE).strftime('%Y-%m-%d--%H:%M:%S')
                logging.error("Weather API: Non 200 response. Check API key, lat, or lon: " + str(response.status_code))
                print(f"{time_now_string} - Weather API: Non 200 response. Check API key, lat, or lon:")
                print(f"{time_now_string} ===> Request status code: ", response.status_code)
                return False
        except Exception as e:
            time_now_string = datetime.now(tz=Config.LOCAL_TIMEZONE).strftime('%Y-%m-%d--%H:%M:%S')
            logging.error("Weather API: Could not get request from the URL. Check API key, lat, or lon: " + str(e))
            print(f"{time_now_string} - Weather API: Could not get request from the URL. Check API key, lat, or lon:")
            print(f"{time_now_string} - ===> " + str(e))
            return False
        return data

def save_weather_data_to_file(filename, data_point):
    try:
        file = Path(filename)
        if file.is_file():
            with open(filename, 'rb') as f:
                file_data = pickle.load(f)
                file_data['time'].append(data_point['time'][0])
                file_data['sampling_time'].append(data_point['sampling_time'][0])
                file_data['temperature'].append(data_point['temperature'][0])
            with open(filename, 'wb') as f:
                pickle.dump(file_data, f)
        else:
            if data_point:
                with open(filename, 'wb') as f:
                    pickle.dump(data_point, f)
    except Exception as e:
        time_now_string = datetime.now(tz=Config.LOCAL_TIMEZONE).strftime('%Y-%m-%d--%H:%M:%S')
        print(f"{time_now_string} - something went wrong writing open_weather_map data to file: ", filename)
        print(str(e))
        return False
    return True

def save_weather_data_to_all_periodic_files(data_point, session_id, time_now):
    try:
        time_now = data_point['time'][0]
    except:
        pass
    fname_list = helpers.get_sensor_data_file_names('open_weather_map', session_id, time_now)
    success_flags = []

    for fname in fname_list:
        if not save_weather_data_to_file(fname, data_point):
            success_flags.append(False)

    if False in success_flags:
        time_now_string = datetime.now(tz=Config.LOCAL_TIMEZONE).strftime('%Y-%m-%d--%H:%M:%S')
        print(f"{time_now_string} - One or more of the weather file saves failed.")
        return False
    return True

########################################################################################




########################################################################################
################################ REQUESTS ##############################################
# These functions are called from the Flask API routes in raspberry-sw/run.py
# and retrieves data from the pickle files, depending on the parameter passed
# to the requests.
#
# A mutex is used to remove the small chance of a bug where requests gets wrong data 
# when several requests are made at the same time. 

def request_data_since(params):
    # params = {
    #   'type':         'hour', 'minute' or 'timestamp'
    #   'date_or_time': an integer, or string timestamp (see readme)
    # }
    # Type: datetime, hour, minute
    #   datetime: alle har samme format! hvilket gjør dette relativt lett
    thread_names = '\t'.join([x.name for x in threading.enumerate()])
    if 'handler-thread' in thread_names:
        global main_q
        global mtx
        mtx.acquire()
        message = ("request_data_since", params, None)
        main_q.put(message)
        data, sensor = request_q.get()
        mtx.release()
    else:
        data, sensor = helpers.request_get_data_since(params)
    return data, sensor

def data_request(params):
    # params = {
    #     'sensor': 'MET', 'tibber', 'tibber-realtime-home-up', 'tibber-realtime-home-pumps', 'sensibo', or 'open_weather_map' 
    #     'format': 'daily'
    #     'date':   'latest', 'first', 'all', '2020-09-02'(daily)
    # }
    thread_names = '\t'.join([x.name for x in threading.enumerate()])
    if 'handler-thread' in thread_names:
        global main_q
        global mtx
        mtx.acquire()
        message = ("data_request", params, None)
        main_q.put(message)
        data, sensor = request_q.get()
        mtx.release()
    else:
        data, sensor = helpers.get_data_for_request(params)
    return data, sensor

def file_list_request(params):
    # params = {
    #     'sensor': 'MET', 'tibber', 'tibber-realtime-home-up', 'tibber-realtime-home-pumps', 'sensibo', 'open_weather_map' 
    #     'format': 'daily'
    #     'date': 'latest', 'first', 'all', '2020-09-02'(daily)
    # }
    try:
        directory = 'data' + os.sep + params['sensor'] + os.sep + params['format']
    except:
        return ['key_error']
    files = helpers.find_files(directory, 'pkl', only_filename=True)
    return files

########################################################################################


########################################################################################
################################ BACKUP LOOP ###########################################

def backup_at_interval(stop_q, interval=Config.BACKUP_INTERVAL):
    time_now = datetime.now(tz=Config.LOCAL_TIMEZONE)

    # Default interval set to 24 hours. Set a semi-random scheduled time to try and not coincide with data collection times.
    scheduled_time = time_now.replace(second=59, microsecond=0, minute=31, hour=21)
    time_now_string = time_now.strftime('%Y-%m-%d--%H:%M:%S')
    print(f"{time_now_string} - Backup started | next scheduled backup: {scheduled_time}")
    
    while True:
        try:
            message = stop_q.get(block=False)
            if message is _sentinel:
                stop_q.put(_sentinel)
                time_now_string = datetime.now(tz=Config.LOCAL_TIMEZONE).strftime('%Y-%m-%d--%H:%M:%S')
                print(f"{time_now_string} - Stopped: Backup")
                break
        except:
            pass
        
        helpers.sleep_until(scheduled_time)

        message = ("backup", None, None)
        main_q.put(message)
        
        scheduled_time += timedelta(hours=interval)






##############################################################################################
################################ MAIN HANDLER LOOP ###########################################
# The main loop which handles data file saving, and data requests from the Flask web API.
# Using FIFO queue and a main loop handler to avoid problems with writing/opening files at the same time

def data_and_request_handler(main_q, request_q, stop_q):
    while True:
        sender, data, session_id = main_q.get()
        time_now = datetime.now(tz=Config.LOCAL_TIMEZONE)
        
        if sender == 'tibber-realtime-home-pumps':
            save_ok = save_tibber_realtime_data_to_all_periodic_files(data_point=data, session_id=session_id, sensor=sender, time_now=time_now)
            helpers.print_save_message(sender, time_now, save_ok)
        elif sender == 'tibber-realtime-home-up':
            save_ok = save_tibber_realtime_data_to_all_periodic_files(data_point=data, session_id=session_id, sensor=sender, time_now=time_now)
            helpers.print_save_message(sender, time_now, save_ok)
        elif sender == 'tibber':
            save_ok = save_tibber_data_to_all_periodic_files(data_point=data, session_id=session_id, time_now=time_now)
            helpers.print_save_message(sender, time_now, save_ok)

        elif sender == 'sensibo':
            save_ok = save_sensibo_data_to_all_periodic_files(data_point=data, session_id=session_id, time_now=time_now)
            helpers.print_save_message(sender, time_now, save_ok)

        elif sender == 'open_weather_map':
            save_ok = save_weather_data_to_all_periodic_files(data_point=data, session_id=session_id, time_now=time_now)
            helpers.print_save_message(sender, time_now, save_ok)
        
        elif sender == 'MET':
            save_ok = save_met_data_to_all_periodic_files(data_point=data, session_id=session_id, time_now=time_now)
            helpers.print_save_message(sender, time_now, save_ok)

        elif helpers.is_api_data_request(sender):
            params = data
            data_, sensor_ = helpers.get_data_for_request(params)
            message = (data_, sensor_)
            request_q.put(message)

        elif sender == 'request_data_since':
            params = data
            data_, sensor_ = helpers.request_get_data_since(params)
            message = (data_, sensor_)
            request_q.put(message)

        elif sender == 'backup':
            backup_ok = backup.backup('data', 'backups')
            helpers.print_save_message(sender, time_now, save_ok)

        elif sender == 'stop' and data is _sentinel:
            stop_q.put(_sentinel)
            time_now_string = datetime.now(tz=Config.LOCAL_TIMEZONE).strftime('%Y-%m-%d--%H:%M:%S')
            print(f'{time_now_string} -------------------------------')
            print(f"{time_now_string} Stopped: The main data_mining handler thread. Waiting for the backup and sensor threads to stop...")
            break

        else:
            time_now_string = datetime.now(tz=Config.LOCAL_TIMEZONE).strftime('%Y-%m-%d--%H:%M:%S')
            print(f"{time_now_string} - Unknown queue request in main handler.")




##############################################################################################
################################ START DATA_COLLECTOR ########################################
def start(params={}):
    # check if the data collector is already running:
    for thread in threading.enumerate():
        if 'handler-thread' in thread.name:
            time_now_string = datetime.now(tz=Config.LOCAL_TIMEZONE).strftime('%Y-%m-%d--%H:%M:%S')
            print(f'{time_now_string} ----------------------------------------------')
            print(f"{time_now_string} Data miner is already started, please stop first.")
            print(f'{time_now_string} ----------------------------------------------')
            return False, "Data_collector is already started"
    
    # Access params. Check .env file if params are missing. Then write new params to .env file.
    # The .env file stores the latest supplied parameters, so that when the software restarts from
    # crashes and disconnects it can just access the latest parameters from the .env file.
    try:
        tibber = params['tibber']
        tibber_api_key = tibber.get('api_key', None)
        tibber_sampling_time = tibber.get('sampling_time', Config.DEFAULT_PARAMS['tibber']['sampling_time'])
        tibber_rt_sampling_time = tibber.get('rt_sampling_time', Config.DEFAULT_PARAMS['tibber']['rt_sampling_time'])
    except:
        time_now_string = datetime.now(tz=Config.LOCAL_TIMEZONE).strftime('%Y-%m-%d--%H:%M:%S')
        print(f"{time_now_string} - No parameters supplied for Tibber, trying to access parameters from .env file")
        tibber_api_key = os.getenv("TIBBER_API_KEY")
        tibber_sampling_time = int(os.getenv("TIBBER_SAMPLING_TIME")) if os.getenv("TIBBER_SAMPLING_TIME") else None
        tibber_rt_sampling_time = int(os.getenv("TIBBER_RT_SAMPLING_TIME")) if os.getenv("TIBBER_RT_SAMPLING_TIME") else None

    try:
        sensibo = params['sensibo']
        sensibo_api_key = sensibo.get('api_key', None)
        sensibo_sampling_time = sensibo.get('sampling_time', Config.DEFAULT_PARAMS['sensibo']['sampling_time'])
    except:
        time_now_string = datetime.now(tz=Config.LOCAL_TIMEZONE).strftime('%Y-%m-%d--%H:%M:%S')
        print(f"{time_now_string} - No parameters supplied for Sensibo, trying to access parameters from .env file")
        sensibo_api_key = os.getenv("SENSIBO_API_KEY")
        sensibo_sampling_time = int(os.getenv("SENSIBO_SAMPLING_TIME")) if os.getenv("SENSIBO_SAMPLING_TIME") else None

    try:
        # The following try/except is to prevent "None" parameters when the data collector is started for the first time and no MET parameters was supplied
        try:
            met = params['MET']
        except:
            if not os.getenv("MET_LAT") or not os.getenv("MET_LON") or not os.getenv("MET_ALTITUDE") or not os.getenv("MET_LOCATION") or not os.getenv("MET_PERSONAL_ID"):
                met = {}
            else:
                raise Exception("Try to access the parameters from .env")
        met_lat = met.get('lat', Config.DEFAULT_PARAMS['MET']['lat'])
        met_lon = met.get('lon', Config.DEFAULT_PARAMS['MET']['lon'])
        met_altitude = met.get('altitude', Config.DEFAULT_PARAMS['MET']['altitude'])
        met_location = met.get('location_name', Config.DEFAULT_PARAMS['MET']['location_name'])
        met_personal_id = met.get('personal_id', Config.DEFAULT_PARAMS['MET']['personal_id'])
    except:
        time_now_string = datetime.now(tz=Config.LOCAL_TIMEZONE).strftime('%Y-%m-%d--%H:%M:%S')
        print(f"{time_now_string} - No parameters supplied for MET, trying to access parameters from .env file")
        met_lat = float(os.getenv("MET_LAT")) if os.getenv("MET_LAT") else None
        met_lon = float(os.getenv("MET_LON")) if os.getenv("MET_LON") else None
        met_altitude = int(os.getenv("MET_ALTITUDE")) if os.getenv("MET_ALTITUDE") else None
        met_location = os.getenv("MET_LOCATION")
        met_personal_id = os.getenv("MET_PERSONAL_ID")

    try:
        weather = params['open_weather_map']
        weather_api_key = weather.get('api_key', None)
        weather_sampling_time = weather.get('sampling_time', Config.DEFAULT_PARAMS['open_weather_map']['sampling_time'])
        weather_lat = weather.get('lat', Config.DEFAULT_PARAMS['open_weather_map']['lat'])
        weather_lon = weather.get('lon', Config.DEFAULT_PARAMS['open_weather_map']['lon'])
    except:
        time_now_string = datetime.now(tz=Config.LOCAL_TIMEZONE).strftime('%Y-%m-%d--%H:%M:%S')
        print(f"{time_now_string} - No parameters supplied for open_weather_map, trying to access parameters from .env file")
        weather_api_key = os.getenv("OPEN_WEATHER_MAP_API_KEY")
        weather_sampling_time = int(os.getenv("OPEN_WEATHER_MAP_SAMPLING_TIME")) if os.getenv("OPEN_WEATHER_MAP_SAMPLING_TIME") else None
        weather_lat = os.getenv("OPEN_WEATHER_MAP_LAT")
        weather_lon = os.getenv("OPEN_WEATHER_MAP_LON")

    # Preparing to write parameters the .env file
    env_params = {}
    env_params['tibber'] = {
        'api_key': tibber_api_key,
        'sampling_time': tibber_sampling_time,
        'rt_sampling_time': tibber_rt_sampling_time
    }
    env_params['sensibo'] = {
        'api_key': sensibo_api_key,
        'sampling_time': sensibo_sampling_time,
    }
    env_params['MET'] = {}
    if met_lat != None:
        env_params['MET']['lat'] = met_lat
    if met_lon != None:
        env_params['MET']['lon'] = met_lon
    if met_altitude != None:
        env_params['MET']['altitude'] = met_altitude 
    if met_location != None:
        env_params['MET']['location_name'] = met_location 
    if met_personal_id != None:
        env_params['MET']['personal_id'] = met_personal_id 
    env_params['open_weather_map'] = {}
    if weather_api_key != None and weather_api_key != 'None':
        env_params['open_weather_map']['api_key'] = weather_api_key
    if weather_sampling_time != None:
        env_params['open_weather_map']['sampling_time'] = weather_sampling_time
    if weather_lat != None:
        env_params['open_weather_map']['lat'] = weather_lat
    if weather_lon != None:
        env_params['open_weather_map']['lon'] = weather_lon

    # Check for tibber and sensibo API keys, then store parameters in .env file and start the data collector
    if not tibber_api_key or not sensibo_api_key or tibber_api_key == 'None' or sensibo_api_key == 'None':
        time_now_string = datetime.now(tz=Config.LOCAL_TIMEZONE).strftime('%Y-%m-%d--%H:%M:%S')
        print(f"{time_now_string} - Tried to start, but missing api keys for Tibber or Sensibo in params or .env file.")
        print("==========")
        return False
    else:
        try:
            helpers.write_to_env_file(env_params)
        except Exception as e:
            time_now_string = datetime.now(tz=Config.LOCAL_TIMEZONE).strftime('%Y-%m-%d--%H:%M:%S')
            print(f"{time_now_string} - Failed to write parameters to .env file: ", e)
        time_now_string = datetime.now(tz=Config.LOCAL_TIMEZONE).strftime('%Y-%m-%d--%H:%M:%S')
        print(f"\n{time_now_string} ====================== STARTING DATA COLLECTOR ===========================")
        print(f"{time_now_string} - Successfully accessed api keys. Starting the data collector.")
        print("==========")

    # generate folder structure for data
    helpers.generate_all_data_folders(['tibber', 'tibber-realtime-home-pumps', 'tibber-realtime-home-up', 'sensibo', 'open_weather_map', 'MET'])

    # initialize Sensor objects, 
    tibber = TibberAPI(
        api_key = tibber_api_key, 
        sampling_time = tibber_sampling_time
        )
    tibberRealtime_Home_Up = TibberRT(
        api_key = tibber_api_key,
        sensor_name = 'tibber-realtime-home-up',
        sensor_number = 0,
        sampling_time = tibber_rt_sampling_time
    )
    tibberRealtime_Home_Pumps = TibberRT(
        api_key = tibber_api_key,
        sensor_name = 'tibber-realtime-home-pumps',
        sensor_number = 1,
        sampling_time = tibber_rt_sampling_time
    )
    sensibo = SensiboSensor(
        api_key = sensibo_api_key, 
        sampling_time = sensibo_sampling_time
        )
    met = METForecast(
        lat             = float(met_lat), 
        lon             = float(met_lon), 
        alti            = int(met_altitude), 
        location_name   = met_location, 
        personal_id     = met_personal_id
    )
    if weather_api_key != None and weather_api_key != 'None':  # Making the open_weather_map sensor thread optional
        weather = WeatherAPI(
            api_key = weather_api_key, 
            lat = weather_lat, 
            lon = weather_lon, 
            sampling_time = weather_sampling_time
        )

    # Global object that signals shutdown
    global _sentinel
    _sentinel = object()

    # Initialize queues. Global so they can be accessed by the Flask api requests.
    global main_q
    global request_q
    global stop_q
    main_q = Queue()
    request_q = Queue()
    stop_q = Queue()

    # Initialize Mutex object --- Used for data requests to avoid a bug where requests 
    # gets the data of another request when requests are placed at the same time and the main 
    # handler threads manages to send back data for both requests within one thread/CPU time unit
    global mtx
    mtx = threading.Lock()

    # start the main handler thread
    main_handler_t = threading.Thread(name='handler-thread', target=data_and_request_handler, args=(main_q, request_q, stop_q))
    main_handler_t.daemon = True
    main_handler_t.start()

    # start each sensor loop thread/process
    tibber_scheduled_time = helpers.get_first_scheduled_time(sampling_time_minutes=tibber_sampling_time)
    tibber_t = threading.Thread(name='tibber-thread', target=tibber.start_mining_loop, args=(main_q, stop_q, tibber_scheduled_time, tibber_sampling_time))
    tibber_t.daemon = True
    tibber_t.start()

    tibber_rt_1_t = threading.Thread(name='tibber-rt-home_up-thread', target=tibberRealtime_Home_Up.start_rt, args=(main_q, stop_q))
    tibber_rt_1_t.daemon = True
    tibber_rt_1_t.start()

    tibber_rt_0_t = threading.Thread(name='tibber-rt-home_pumps-thread', target=tibberRealtime_Home_Pumps.start_rt, args=(main_q, stop_q))
    tibber_rt_0_t.daemon = True
    tibber_rt_0_t.start()

    sensibo_scheduled_time = helpers.get_first_scheduled_time(sampling_time_minutes=sensibo_sampling_time)
    sensibo_t = threading.Thread(name='sensibo-thread', target=sensibo.start_mining_loop, args=(main_q, stop_q, sensibo_scheduled_time, sensibo_sampling_time))
    sensibo_t.daemon = True
    sensibo_t.start()

    met_t = threading.Thread(name='met-thread', target=met.start_mining_loop, args=(main_q, stop_q))
    met_t.daemon = True
    met_t.start()

    if weather_api_key != None and weather_api_key != 'None':  # Making the open_weather_map sensor thread optional
        weather_scheduled_time = helpers.get_first_scheduled_time(sampling_time_minutes=weather_sampling_time)
        weather_t = threading.Thread(name='open-weather-map-thread', target=weather.start_mining_loop, args=(main_q, stop_q, weather_scheduled_time, weather_sampling_time))
        weather_t.daemon = True
        weather_t.start()

    # start backup thread
    if params:
        backup = params.get('backup', Config.DEFAULT_PARAMS['backup'])
    else:
        backup = Config.DEFAULT_PARAMS['backup']
    if backup['run_backups']:
        backup_t = threading.Thread(name='backup-thread', target=backup_at_interval, args=(stop_q, Config.BACKUP_INTERVAL))
        backup_t.daemon = True
        backup_t.start()

    # check if the program is running in a webserver
    if params:
        running_in_webserver = params.get('running_in_webserver', Config.DEFAULT_PARAMS['running_in_webserver'])
    else:
        running_in_webserver = Config.DEFAULT_PARAMS['running_in_webserver'] 
    if not running_in_webserver:
        time_now_string = datetime.now(tz=Config.LOCAL_TIMEZONE).strftime('%Y-%m-%d--%H:%M:%S')
        print(f"{time_now_string} - Waiting for threads")
        tibber_t.join()
        tibber_rt_0_t.join()
        tibber_rt_1_t.join()
        sensibo_t.join()
        if weather_api_key != None and weather_api_key != 'None': # Making the open_weather_map sensor thread optional
            weather_t.join()
        met_t.join()
        main_handler_t.join()
        return True, "started (not on webserver)"
    return True, "Data_collector started on webserver."



##############################################################################################
################################ STOP DATA_COLLECTOR #########################################
def stop():
    global main_q
    global _sentinel
    message = ('stop', _sentinel, None)
    main_q.put(message)
    return "Stopping the data collector"
