from pathlib import Path
import zipfile
from datetime import datetime
import os

from data_collector.helpers import *
from data_collector.config import Config

logging.basicConfig(filename='log__data_collector.log',
                    filemode='a',
                    format='%(asctime)s,%(msecs)d %(name)s %(levelname)s %(message)s',
                    datefmt='%Y-%m-%d--%H:%M:%S',
                    level=logging.WARNING)

def backup(object_to_backup, backup_folder):
    # settings
    max_backup_copies = Config.BACKUP_COPIES

    # Setup paths and folder
    object_to_backup_path = Path(object_to_backup)
    backup_directory_path = Path(backup_folder)
    if not object_to_backup_path.exists():
        print("Backup: object does not exist")
        logging.error("Backup: object does not exist")
        return False
    backup_directory_path.mkdir(parents=True, exist_ok=True)

    # check how many existing backups exist
    existing_backups = [
        x for x in backup_directory_path.iterdir()
        if x.is_file() and x.suffix == '.zip' and x.name.startswith('data_backup')
    ]

    # Enforce max backups and delete oldest if there will be too many after the new backup
    oldest_to_newest_backup_by_name = list(sorted(existing_backups, key=lambda f: f.name))
    while len(oldest_to_newest_backup_by_name) >= max_backup_copies:  # >= because we will have another soon
        backup_to_delete = oldest_to_newest_backup_by_name.pop(0)
        backup_to_delete.unlink()

    # Create zip file (for both file and folder options)
    backup_file_name = f'{object_to_backup_path.name}_backup__{datetime.now(tz=Config.LOCAL_TIMEZONE).strftime("%Y-%m-%d_%H-%M-%S")}.zip'
    zip_file = zipfile.ZipFile(str(backup_directory_path / backup_file_name), mode='w', strict_timestamps=False)
    if object_to_backup_path.is_file():
        # If the object to write is a file, write the file
        zip_file.write(
            object_to_backup_path.absolute(),
            arcname=object_to_backup_path.name,
            compress_type=zipfile.ZIP_DEFLATED
        )
    elif object_to_backup_path.is_dir():
        # If the object to write is a directory, write all the files
        for file in object_to_backup_path.glob('**/*'):
            try:
                if file.is_file() and os.stat(file.absolute()).st_size != 0:
                    zip_file.write(
                        file.absolute(),
                        arcname=str(file.relative_to(object_to_backup_path))#,
                        # compress_type=zipfile.ZIP_DEFLATED
                    )
            except Exception as e:
                logging.error("Backup:")
                logging.error(str(e))
                return False
    zip_file.close()
    return True