from flask import Flask
from flask import jsonify, request
import time
from flask.json import JSONEncoder
import data_collector

app = Flask(__name__)

# Start the data collector automatically when the software has been restarted
def autostart_data_collector():
    print("=== Trying to Autostart Data Collector ===")
    started = data_collector.start()
autostart_data_collector()


######################### The routes of the web api

@app.route('/')
@app.route('/index')
def index():
  return """
            <xmp>
            Static files:
            ==============================
            /readme     : README.md
            /data/      : download static .pkl data files 
            /backups/   : download .zip backups
            

            API calls:
            ==============================
            /api/get_file_list/<sensor>/<date_format>/<date>    : Gets a list of all files (returns json)
            /api/get_data/<sensor>/<date_format>/<date>         : Gets data (returns json)
                <sensor>        : 'tibber', 
                                  'tibber-realtime-home-pumps', 
                                  'tibber-realtime-home-up', 
                                  'sensibo', 
                                  'MET',
                                  'open_weather_map'
                <date_format>   : 'daily'
                <date>          : 'first',      (Gets the data from the first data file that was created)
                                  'latest',     (Gets the data from the latest data file that was created)
                                  'all',        (Gets all data from all data files)
                                  '2020-09-02', (Gets all data from a specific day)
            
            /api/get_data_since/<sensor>/<time_type>/<time>     : Gets data since a specified time (returns json)
                <sensor>        : (same)
                <time_type>     : 'hour'
                                  'minute'
                                  'timestamp'
                <time>          :  e.g. '5' (a positive integer, if <time_type> is 'hour' or 'minute')
                                   e.g. 'Wed, 09 Jun 2021 11:49:09 GMT' (a string timestamp, if <time_type> is 'timestamp')
        
            Starting or stopping the data_collector: see README.md
            </xmp>
         """

@app.route('/api/start_data_collector', methods=['POST'])
def start_data_collector():
    if request.method == 'POST':
        params = request.get_json()

        if params:
            response, status_code = check_params(params)
            if status_code != 200:
                return jsonify(response), status_code
        
        # If params are ok, or no params were supplied, try to start the data_collector
        started, message = data_collector.start(params)
        response['message'] = message
        if started:
            return jsonify(response), 200
        else:
            response['status_code'] = 500
            return jsonify(response), 500
    else:
        return "Wrong request type. Needs POST request.", 400

@app.route('/api/stop_data_collector', methods=['POST'])
def stop_data_collector():
    if request.method == 'POST':
        data_collector.stop()
        response = {
                'message': 'Data collector stopped.',
                'status_code': 200
            }
        return jsonify(response), 200
    else:
        return "Wrong request type. Need POST request.", 400

@app.route('/api/get_file_list/<sensor>/<date_format>/<date>', methods=['GET'])
def get_file_list(sensor, date_format, date):
    if request.method == 'GET':
        params = {
            'sensor': sensor,
            'format': date_format,
            'date': date
        }
        file_list = data_collector.file_list_request(params)
        if not file_list:
            file_list = None
            status_code = 400
        else:
            status_code = 200
        response = {
            'params': params,
            'files': file_list,
            'status_code': status_code
        }
        return jsonify(response), 200
    else:
        return "Wrong request. Needs GET request.", 400

@app.route('/api/get_data/<sensor>/<date_format>/<date>', methods=['GET'])
def get_data(sensor, date_format, date):
    if request.method == 'GET':
        params = {
            'sensor': sensor,
            'format': date_format,
            'date': date
        }
        data, sensor = data_collector.data_request(params)

        if data == False or sensor == False:
            return "bad request", 400

        response = {
            'sensor': sensor,
            'data': data,
            'status_code': 200
        }
        return jsonify(response), 200
    else:
        return "Wrong request. Needs GET request.", 400

@app.route('/api/get_data_since/<sensor>/<type>/<date_or_time>', methods=['GET'])
def get_data_since(sensor, type, date_or_time):
    # TODO
    # Type: timestamp, hours, minutes
    #   datetime: alle har samme format! hvilket gjør dette relativt lett

    if request.method == 'GET':
        params = {
            'sensor': sensor,
            'type': type,
            'date_or_time': date_or_time
        }
        data, sensor_or_msg = data_collector.request_data_since(params)

        if data == False:
            return sensor_or_msg, 400

        response = {
            'sensor': sensor_or_msg,
            'data': data,
            'status_code': 200
        }
        return jsonify(response), 200
    else:
        return "Wrong request. Needs GET request.", 400
    return

@app.route('/api/get_latest_data_point/<sensor>', methods=['GET'])
def get_latest_data_point(sensor):
    # TODO
    return

# Todo later
@app.route('/api/control', methods=['GET', 'POST'])
def control():
  if request.method == 'GET':
    pass
  elif request.method == 'POST':
    pass
  else:
    pass
  return "Control: to be implemented"

def check_params(params):
    keys = params.keys()
    if 'tibber' not in keys or 'sensibo' not in keys:
        response = {
            'error': 'key_error',
            'message': "Needs to add all sensors to params: tibber, sensibo, open_weather_map",
            'status_code': 400
        }
        return response, 400
    sensors = ['tibber', 'sensibo']
    for sensor in sensors:
        if not params[sensor].get('api_key', False):
            response = {
                'error': 'key_error',
                'message': "missing api_key for: " + sensor,
                'status_code': 400
            }
            return response, 400
        if params[sensor].get('sampling_time', 1) <= 0 or type(params[sensor].get('sampling_time', 1)) != int \
            or params[sensor].get('rt_sampling_time', 1) <=0 or type(params[sensor].get('rt_sampling_time', 1)) != int:
            response = {
                'error': 'type_error',
                'message': "sampling time needs to be an integer > 0: " + sensor,
                'status_code': 400
            }
            return response, 400
    if params.get('backup') and type(params['backup'].get('run_backups')) != bool:
        response = {
            'error': 'type_error',
            'message': "run_backups must be a bool: " + sensor,
            'status_code': 400
        }
        return response, 400
    
    # If no errors in params:
    response = {
        'message': '',
        'params': params,
        'status_code': 200
    }
    return response, 200